"""Generic data types for information retrieval"""

from abc import ABC, abstractmethod
from functools import cached_property
from pathlib import Path
from attrs import define
from typing import Callable, Dict, Iterator, List, Optional, Tuple, Type
import random
from experimaestro import Config
from datamaestro.definitions import datatasks, Param, Meta
from dataclasses import dataclass
from datamaestro.data import Base
from datamaestro_text.utils.files import auto_open
from .base import (
    Document,
    Topic,
    AdhocAssessment,
    IDTopic,
    IDDocument,
    TextTopic,
    TextDocument,
)


class Documents(Base):
    """A set of documents with identifiers

    See `IR Datasets <https://ir-datasets.com/index.html>`_ for the list of query classes
    """

    count: Meta[Optional[int]]
    """Number of documents"""

    def iter(self) -> Iterator[Document]:
        """Returns an iterator over documents"""
        raise self.iter_documents()

    def iter_documents(self) -> Iterator[Document]:
        return self.iter()

    def iter_ids(self) -> Iterator[str]:
        """Iterates over document ids

        By default, use iter_documents, which is not really efficient.
        """
        for doc in self.iter():
            yield doc.get_id()

    @property
    def documentcount(self):
        """Returns the number of terms in the index"""
        if self.count is not None:
            return self.count

        raise NotImplementedError(f"For class {self.__class__}")

    @property
    def document_cls(self) -> Type[Document]:
        """The class for documents"""
        raise NotImplementedError(f"For class {self.__class__}")


class DocumentStore(Documents):
    """A document store

    A document store can
    - match external/internal ID
    - return the document content
    - return the number of documents
    """

    def docid_internal2external(self, docid: int):
        """Converts an internal collection ID (integer) to an external ID"""
        raise NotImplementedError(f"For class {self.__class__}")

    def document_int(self, internal_docid: int) -> Document:
        """Returns a document given its internal ID"""
        docid = self.docid_internal2external(internal_docid)
        return self.document(docid)

    def document_ext(self, docid: str) -> Document:
        """Returns a document given its external ID"""
        raise NotImplementedError(f"document() in {self.__class__}")

    def iter_sample(
        self, randint: Optional[Callable[[int], int]]
    ) -> Iterator[Document]:
        """Sample documents from the dataset"""
        length = self.count()
        randint = randint or (lambda max: random.randint(0, max - 1))
        while True:
            yield self.document_int(randint(length))


class AdhocIndex(DocumentStore):
    """An index can be used to retrieve documents based on terms"""

    @property
    def termcount(self):
        """Returns the number of terms in the index"""
        raise NotImplementedError(f"For class {self.__class__}")

    def term_df(self, term: str):
        """Returns the document frequency"""
        raise NotImplementedError(f"For class {self.__class__}")


class Topics(Base):
    """A set of topics with associated IDs"""

    def iter(self) -> Iterator[Topic]:
        """Returns an iterator over topics"""
        raise NotImplementedError(f"For class {self.__class__}")

    def count(self) -> Optional[int]:
        """Returns the number of topics if known"""
        return None

    @property
    def topic_cls(self) -> Type[Topic]:
        """The class for topics"""
        raise NotImplementedError(f"For class {self.__class__}")


AdhocTopics = Topics


class TopicsStore(Topics):
    """Adhoc topics store"""

    @abstractmethod
    def topic_int(self, internal_topic_id: int) -> Topic:
        """Returns a document given its internal ID"""

    @abstractmethod
    def topic_ext(self, external_topic_id: int) -> Topic:
        """Returns a document given its external ID"""


class AdhocAssessments(Base, ABC):
    """Ad-hoc assessments (qrels)"""

    def iter(self) -> Iterator[AdhocAssessment]:
        """Returns an iterator over assessments"""
        raise NotImplementedError(f"For class {self.__class__}")


class AdhocRun(Base):
    """IR adhoc run"""

    pass


class AdhocResults(Base):
    def get_results(self) -> Dict[str, float]:
        """Returns the aggregated results

        :return: Returns a dictionary where each metric (keys) is associated
            with a value
        """
        raise NotImplementedError(f"For class {self.__class__}")


@datatasks("information retrieval")
class Adhoc(Base):
    """An Adhoc IR collection with documents, topics and their assessments"""

    documents: Param[Documents]
    """The set of documents"""

    topics: Param[Topics]
    """The set of topics"""

    assessments: Param[AdhocAssessments]
    """The set of assessments (for each topic)"""


class RerankAdhoc(Adhoc):
    """Re-ranking ad-hoc task based on an existing run"""

    run: Param[AdhocRun]
    """The run to re-rank"""


class Measure(Config):
    """An Information Retrieval measure"""

    pass


class TrainingTriplets(Base):
    """Triplet for training IR systems: query / query ID, positive document,
    negative document"""

    def iter(self) -> Iterator[Tuple[Topic, Document, Document]]:
        raise NotImplementedError(f"For class {self.__class__}")

    def count(self):
        """Returns the number of triplets or None"""
        return None

    @property
    def topic_cls(self) -> Type[Topic]:
        """The class for topics"""
        raise NotImplementedError(f"For class {self.__class__}")

    @property
    def document_cls(self) -> Type[Document]:
        """The class for documents"""
        raise NotImplementedError(f"For class {self.__class__}")


class TrainingTripletsLines(TrainingTriplets):
    """Training triplets with one line per triple (query texts)"""

    sep: Meta[str]
    path: Param[Path]

    doc_ids: Meta[bool]
    """True if we have documents IDs"""

    topic_ids: Meta[bool]
    """True if we have query IDs"""

    def iter(self) -> Iterator[Tuple[Topic, Document, Document]]:
        with auto_open(self.path, "rt") as fp:
            for line in fp:
                q, pos, neg = line.strip().split(self.sep)
                yield self.topic_cls(q), self.document_cls(pos), self.document_cls(neg)

    @cached_property
    def topic_cls(self) -> Type[Topic]:
        """The class for topics"""
        return IDTopic if self.topic_ids else TextTopic

    @cached_property
    def document_cls(self) -> Type[Document]:
        """The class for documents"""
        return IDDocument if self.doc_ids else TextDocument


@define(kw_only=True)
class PairwiseSample:
    """A a query with positive and negative samples"""

    topic: Topic
    """The topic"""

    positives: List[Document]
    """Relevant documents"""

    negatives: Dict[str, List[Document]]
    """Non relevant documents, organized in a dictionary where keys
    are the algorithm used to retrieve the negatives"""

    @property
    def topic_cls(self):
        """The class for topics"""
        raise NotImplementedError(f"For class {self.__class__}")

    @property
    def document_cls(self):
        """The class for documents"""
        raise NotImplementedError(f"For class {self.__class__}")


class PairwiseSampleDataset(Base):
    """Datasets where each record is a query with positive and negative samples"""

    def iter(self) -> Iterator[PairwiseSample]:
        raise NotImplementedError(f"For class {self.__class__}")



from typing import Iterator
from experimaestro import Meta
from datamaestro.data.huggingface import HuggingFaceDataset
from . import PairwiseSample, PairwiseSampleDataset


class HuggingFacePairwiseSampleDataset(HuggingFaceDataset, PairwiseSampleDataset):
    """Triplet for training IR systems: query / query ID, positive document, negative document

    Attributes:

        ids: True if the triplet is made of IDs, False otherwise
    """

    ids: Meta[bool]

    query_id: Meta[str] = "qid"
    """The name of the field containing the query ID"""

    pos_id: Meta[str] = "pos"
    """The name of the field containing the positive samples"""

    neg_id: Meta[str] = "neg"
    """The name of the field containing the negative samples"""

    def iter(self) -> Iterator[PairwiseSample]:
        for element in self.data:
            yield PairwiseSample(
                element[self.query_id], element[self.pos_id], element[self.neg_id]
            )



"""CORD-19 dataset
"""

from datamaestro.annotations.agreement import useragreement
from datamaestro.definitions import datatasks, dataset
from datamaestro.download import reference
from datamaestro.download.single import filedownloader
from datamaestro.utils import HashCheck
from datamaestro_text.data.ir import Adhoc
import datamaestro_text.data.ir.cord19 as d_cord19
from datamaestro_text.data.ir.trec import TrecAdhocAssessments

cord19_lua = useragreement(
    """COVID DATASET LICENSE AGREEMENT

By accessing, downloading or otherwise using any Journals, Articles, Metadata, Abstracts,
Full-Texts or any other content types provided in the COVID-19 Open Research Dataset (CORD-19)
Database (the “Data”), You expressly acknowledge and agree to the following:

• AI2 grants to You a worldwide, perpetual, non-exclusive, non-transferablelicenseto use and
make derivatives of the Datafor text and data mining only.

• AI2 warrants that it has the right to make the Data available to Youas provided for in and
subject to this Agreement and in accordance with applicable law.  EXCEPT FOR THE LIMITED WARRANTY
IN THIS SECTION, THE DATA IS PROVIDED “AS IS”, WITHOUT ANY WARRANTIES OF ANY KIND.

• You agree to comply with all applicable local, state, national, and international laws and
regulations with respect to AI2’s license and Youruse of the Data.• Data provided by AI2 is
from copyrighted sources of the respective copyright holders. You are solely responsible
for Your and Your users’ compliance with any copyright, patent or trademark restrictions
and are referred to the copyright, patent or trademark notices appearing in the original
sources, all of which are hereby incorporated by reference""",
    id="ai2.cord19",
)


@cord19_lua
@filedownloader(
    "data.csv",
    url="https://ai2-semanticscholar-cord-19.s3-us-west-2.amazonaws.com/2020-07-16/metadata.csv",
    checker=HashCheck("80d664e496b8b7e50a39c6f6bb92e0ef"),
)
@dataset(
    d_cord19.Documents,
    url="https://ir.nist.gov/covidSubmit/index.html",
)
def cord19_round5_metadata(data):
    """Cord 19 metadata (round 5)

    Released on 2020-07-16
    """
    return {
        "path": data,
        "names_row": 0,
        # Number of documents
        "count": 192509,
    }


@filedownloader(
    "data.xml",
    url="https://ir.nist.gov/covidSubmit/data/topics-rnd5.xml",
    checker=HashCheck("0307a37b6b9f1a5f233340a769d538ea"),
)
@dataset(d_cord19.Topics)
def cord19_round5_topics(data):
    """CORD-19 topics (round 5)"""
    return {"path": data}


@filedownloader(
    "data.ssv",
    url="https://ir.nist.gov/covidSubmit/data/qrels-covid_d5_j0.5-5.txt",
    checker=HashCheck("8138424a59daea0aba751c8a891e5f54"),
)
@dataset(TrecAdhocAssessments)
def cord19_round5_assessments(data):
    """CORD19 assessments (round 5)"""
    return {"path": data}


@reference("collection", cord19_round5_metadata)
@reference("topics", cord19_round5_topics)
@reference("qrels", cord19_round5_assessments)
@datatasks("information retrieval", "passage retrieval")
@dataset(Adhoc, url="https://ir.nist.gov/covidSubmit/data.html")
def cord19_round5(topics, qrels, collection):
    """CORD-19 IR collection (round 5)

    This is the primary test collection for ad hoc retrieval that is the outcome of all five rounds of TREC-COVID. The test set, called TREC-COVID Complete, consists of the Round 5 document set (July 16 release of CORD-19); the final set of 50 topics; and the cumulative judgments from all assessing rounds with CORD-UIDs mapped to July 16 ids if necessary, previously judged documents no longer in the July 16 release removed, and the last judgments for documents judged multiple times due to significant content changes between rounds. Note that no TREC-COVID submissions correspond to this collection since all TREC-COVID submissions were subject to residual collection evaluation.
    """
    return {
        "documents": collection,
        "topics": topics,
        "assessments": qrels,
    }



from abc import ABC, abstractmethod
from typing import Generic, Iterator, List, Optional, TypeVar, get_origin, get_args
from attr import define
from datamaestro.data import Base


@define(kw_only=True, slots=False)
class Entry:
    query: str
    """The query issued by the user"""


@define(kw_only=True, slots=False)
class DecontextualizedEntry(Entry):
    decontextualized_query: str
    """Human rewritten query"""


@define(kw_only=True, slots=False)
class AnswerEntry(Entry):
    answer: str
    """The system answer"""


@define(kw_only=True, slots=False)
class RetrievedEntry(Entry):
    """List of retrieved documents and their relevance"""

    documents: List[str]
    """List of retrieved documents"""

    document_relevances: Optional[List[str]] = None
    """List of retrieved documents and their relevance status"""


GenericEntry = TypeVar("GenericEntry", bound=Entry)


@define(kw_only=True)
class Conversation(Generic[GenericEntry]):
    id: str
    """The conversation ID"""

    history: List[GenericEntry]
    """The history"""


class ConversationDataset(Base, ABC, Generic[GenericEntry]):
    """A dataset composed of"""

    @classmethod
    def entry_cls(cls, origin=True):
        """Return the generic type class of the entry

        :param origin: if True, return the origin class, otherwise returns a generic type
        """
        hint = next(
            c
            for c in cls.__orig_bases__
            if issubclass(get_origin(c), ConversationDataset)
        )
        arg = get_args(hint)[0]
        return get_origin(arg) if origin else arg

    @abstractmethod
    def iter_conversations(self) -> Iterator[Conversation[Entry]]:
        """Return an iterator over conversations"""
        ...



from attrs import define
from typing import ClassVar, List


class BaseHolder:
    """Base class for topics and documents"""

    has_id: ClassVar[bool] = False
    has_text: ClassVar[bool] = False

    def get_text(self):
        raise RuntimeError(
            f"{type(self)} cannot extract a single text: " "you should use an adapter"
        )

    def get_id(self):
        raise RuntimeError(f"{type(self)} has no ID: " "you should use an adapter")


@define()
class IDHolder:
    """Base data class for ID only data structures"""

    id: str
    has_id: ClassVar[bool] = True

    def get_id(self):
        return self.id


@define()
class TextHolder:
    """Base data class for text only data structures"""

    text: str
    has_text: ClassVar[bool] = True

    def get_text(self):
        return self.text


@define()
class TextAndIDHolder:
    """Base data class for ID and text data structures"""

    id: str
    text: str

    has_id: ClassVar[bool] = True
    has_text: ClassVar[bool] = True

    def get_id(self):
        return self.id

    def get_text(self):
        return self.text


class Document(BaseHolder):
    """Base class for documents"""

    pass


@define()
class TextDocument(TextHolder, Document):
    """Documents with text"""


@define()
class IDDocument(IDHolder, Document):
    """Documents with ID"""


@define()
class GenericDocument(TextAndIDHolder, Document):
    """Documents with ID and text"""


class Topic(BaseHolder):
    pass


@define()
class GenericTopic(TextAndIDHolder, Topic):
    pass


@define()
class TextTopic(TextHolder, Topic):
    pass


@define()
class IDTopic(IDHolder, Topic):
    pass


@define()
class AdhocAssessment:
    doc_id: str


@define()
class SimpleAdhocAssessment(AdhocAssessment):
    rel: float
    """Relevance (> 0 if relevant)"""


@define()
class AdhocAssessedTopic:
    topic_id: str
    """The topic ID"""

    assessments: List[AdhocAssessment]
    """List of assessments for this topic"""



import inspect
from typing import Any, Callable, Dict, Type


class Handler:
    """Returns a result that depends on the type of the argument

    Example:
    ```
    handler = Handler()

    @handler()
    def trectopics(topics: TrecTopics):
        return ("-topicreader", "Trec", "-topics", topics.path)

    @handler()
    def tsvtopics(topics: ir_csv.Topics):
        return ("-topicreader", "TsvInt", "-topics", topics.path)

    command.extend(handler[topics])

    ```
    """

    def __init__(self):
        self.handlers: Dict[Type, Callable[[Any], None]] = {}
        self.defaulthandler = None

    def default(self):
        assert self.defaulthandler is None

        def annotate(method):
            self.defaulthandler = method
            return method

        return annotate

    def __call__(self):
        def annotate(method):
            spec = inspect.getfullargspec(method)
            assert len(spec.args) == 1 and spec.varargs is None

            self.handlers[spec.annotations[spec.args[0]]] = method

        return annotate

    def __getitem__(self, key: object):
        handler = self.handlers.get(key.__class__, None)
        if handler is None:
            if self.default is None:
                raise RuntimeError(
                    f"No handler for {key.__class__} and no default handler"
                )
            handler = self.defaulthandler

        return handler(key)



from typing import Dict, List, Optional
from datamaestro.definitions import argument, Option
from datamaestro.data import Base
from experimaestro import documentation, Param, Config
from pathlib import Path
from datamaestro_text.data.ir import (
    Documents,
    Topics,
    AdhocAssessments,
    AdhocRun,
    AdhocResults,
    Measure,
)


class TrecTopics(Topics):
    path: Option[Path]
    parts: Option[List[str]]

    @documentation
    def iter(self):
        """Iterate over TREC adhoc topics"""
        import datamaestro_text.interfaces.trec as trec

        yield from trec.parse_query_format(self.path)


class TrecAdhocAssessments(AdhocAssessments):
    path: Option[Path]

    def trecpath(self):
        return self.path

    @documentation
    def iter(self):
        """Iterate over TREC adhoc topics"""
        import datamaestro_text.interfaces.trec as trec

        yield from trec.parse_qrels(self.path)


@argument("path", type=Path)
class TrecAdhocRun(AdhocRun):
    pass


class TrecAdhocResults(AdhocResults):
    """Adhoc results (TREC format)"""

    metrics: Param[List[Measure]]
    """List of reported metrics"""

    results: Param[Path]
    """Main results"""

    detailed: Param[Optional[Path]]
    """Results per topic (if any)"""

    def get_results(self) -> Dict[str, float]:
        """Returns the results as a dictionary {metric_name: value}"""
        import re

        re_spaces = re.compile(r"\s+")

        results = {}
        with self.results.open("rt") as fp:
            for line in fp.readlines():
                metric, _, value = re_spaces.split(line.strip())
                results[metric] = value

        return results


class TipsterCollection(Documents):
    path: Param[Path]



from contextlib import contextmanager
import os
from typing import Iterable
from ir_datasets import registry, corpus_id, load

from datamaestro.definitions import AbstractDataset
from .data import (
    AdhocAssessments,
    AdhocRun,
    Documents,
    Topics,
    Adhoc,
    IRDSId,
    TrainingTriplets,
)


class Dataset(AbstractDataset):
    SUFFIX = ""
    __configtype__ = None

    def __init__(self, repository, irds_id, irds_ds):
        super().__init__(repository)
        self.id = (
            f"""irds.{irds_id.replace("/", ".")}"""
            f"""{"." + self.SUFFIX if self.SUFFIX else ""}"""
        )
        self.irds_id = irds_id
        self.irds_ds = irds_ds

    @property
    def fullid(self):
        return f"{self.id}@{self.repository.name}"

    @property
    def description(self):
        return self.irds_ds.documentation()["desc"]

    def hasfiles(self):
        return False

    def download(self, force=False):
        return True

    def prepare(self, download=True):
        ds: IRDSId = super().prepare(download=download)
        ds.irds = self.irds_id
        if download:
            self.download()

        return ds

    @property
    def configtype(self):
        return self.__class__.__configtype__


class QrelsDataset(Dataset):
    SUFFIX = "qrels"
    configtype = AdhocAssessments

    def download(self, force=False):
        # Triggers download
        next(self.irds_ds.qrels_iter())
        return True

    def _prepare(self, download=False) -> Documents:
        return AdhocAssessments(id=self.fullid)


class QueriesDataset(Dataset):
    SUFFIX = "queries"
    configtype = Topics

    def download(self, force=False):
        # Triggers download
        next(self.irds_ds.queries_iter())
        return True

    def _prepare(self, download=False) -> Documents:
        return Topics(id=self.fullid)


# class ScoredDocuments(Dataset):
#     SUFFIX = "scored-documents"
#     base = ScoredDocuments


class DocumentsDataset(Dataset):
    SUFFIX = "documents"
    configtype = Documents

    def download(self, force=False):
        # Triggers download
        next(self.irds_ds.docs_iter())
        return True

    def _prepare(self, download=False) -> Documents:
        return Documents(id=self.fullid)


class TrainingTripletsDataset(Dataset):
    SUFFIX = "docpairs"

    def _prepare(self, download=False) -> Documents:
        return TrainingTriplets(
            id=self.fullid,
        )


class AdhocRunDataset(Dataset):
    SUFFIX = "scoreddocs"
    base = AdhocRun
    configtype = AdhocRun

    def _prepare(self, download=False) -> Documents:
        return AdhocRun(id=self.fullid)


class Collection(Dataset):
    base = Adhoc
    assessements: QrelsDataset
    topics: QueriesDataset

    def _prepare(self, download=False) -> Documents:
        return Adhoc(
            id=self.fullid,
            topics=self.topics.prepare(download),
            assessments=self.assessments.prepare(download),
            documents=self.documents.prepare(download),
        )


class Datasets:
    """Simple wrapper holding related data pieces"""

    def __init__(self, key, title, description):
        self.id = key
        self.title = title
        self.description = description
        self.datasets = []

    def __iter__(self) -> Iterable[AbstractDataset]:
        return self.datasets.__iter__()


IRDS_NO_WARNING_KEY = "IR_DATASETS_SKIP_DEPRECATED_WARNING"


@contextmanager
def no_deprecated_warnings():
    """Context manager to remove ir dataset warnings"""
    old = os.environ.get(IRDS_NO_WARNING_KEY, "")
    os.environ[IRDS_NO_WARNING_KEY] = "true"
    yield None
    os.environ[IRDS_NO_WARNING_KEY] = old


def build(repository):
    """Builds a datamaestro repository by using ir_datasets registry"""
    datasets = {}
    bykey = {}

    def add(cid, ds):
        datasets[cid].datasets.append(ds)
        bykey[ds.id] = ds

    with no_deprecated_warnings():
        for dataset_id in registry:
            ds = load(dataset_id)

            # Skip deprecated datasets
            if hasattr(ds, "deprecated"):
                continue

            if not ds.has_docs():
                # Abstract dataset
                continue

            cid = corpus_id(dataset_id)
            queries = None
            qrels = None

            if cid == dataset_id:
                # If the corpus ID is the current dataset ID
                module = Datasets(
                    cid,
                    ds.documentation().get("pretty_name", cid),
                    ds.documentation()["desc"],
                )
                datasets[cid] = module
                add(cid, DocumentsDataset(repository, dataset_id, ds))

            if ds.has_queries():
                queries = QueriesDataset(repository, dataset_id, ds)
                add(cid, queries)

            if ds.has_docpairs():
                add(cid, TrainingTripletsDataset(repository, dataset_id, ds))

            if ds.has_scoreddocs():
                add(cid, AdhocRunDataset(repository, dataset_id, ds))

            if ds.has_qrels():
                qrels = QrelsDataset(repository, dataset_id, ds)
                add(cid, qrels)

            if qrels and queries:
                collection = Collection(repository, dataset_id, ds)
                collection.documents = datasets[cid].datasets[0]
                collection.topics = queries
                collection.assessments = qrels

                add(cid, collection)

        return list(datasets.values()), bykey



import logging
from typing import Any, Iterator, Tuple, Type
import attrs
import ir_datasets
from ir_datasets.formats import GenericDoc, GenericQuery, GenericDocPair
import ir_datasets.datasets as _irds
from experimaestro import Config
from experimaestro.compat import cached_property
from experimaestro import Option
import datamaestro_text.data.ir as ir
from datamaestro_text.data.ir.base import (
    Topic,
    Document,
    GenericDocument,
    GenericTopic,
    AdhocAssessedTopic,
    SimpleAdhocAssessment,
    IDDocument,
    IDTopic,
)
import datamaestro_text.data.ir.formats as formats


# Interface between ir_datasets and datamaestro:
# provides adapted data types


class IRDSId(Config):
    irds: Option[str]
    """The id to load the dataset from ir_datasets"""

    @classmethod
    def __xpmid__(cls):
        return f"ir_datasets.{cls.__qualname__}"

    @cached_property
    def dataset(self):
        return ir_datasets.load(self.irds)

    def iter(self) -> Iterator[ir.Topic]:
        """Returns an iterator over topics"""
        for query in self.dataset.queries_iter():
            yield self.factory(query)

    def count(self):
        return self.dataset.queries_count()


class AdhocAssessments(ir.AdhocAssessments, IRDSId):
    def iter(self):
        """Returns an iterator over assessments"""
        ds = self.dataset

        class Qrels(dict):
            def __missing__(self, key):
                qqrel = AdhocAssessedTopic(key, [])
                self[key] = qqrel
                return qqrel

        qrels = Qrels()
        for qrel in ds.qrels_iter():
            qrels[qrel.query_id].assessments.append(
                SimpleAdhocAssessment(qrel.doc_id, qrel.relevance)
            )

        return qrels.values()


class tuple_constructor:
    def __init__(self, target_cls: Type, *fields: str):
        self.target_cls = target_cls
        self.fields = fields

    def check(self, source_cls: Type):
        assert (
            source_cls._fields == self.fields
        ), f"Internal error: Fields do not match ({source_cls._fields} and {self.fields})"

    def __call__(self, entry):
        return self.target_cls(*tuple(entry))


@attrs.define()
class IRDSDocumentWrapper(ir.Document):
    doc: Any


class Documents(ir.DocumentStore, IRDSId):
    CONVERTERS = {
        GenericDoc: tuple_constructor(GenericDocument, "doc_id", "text"),
        _irds.beir.BeirCordDoc: tuple_constructor(
            formats.CordDocument, "doc_id", "text", "title", "url", "pubmed_id"
        ),
    }

    """Wraps an ir datasets collection -- and provide a default text
    value depending on the collection itself"""

    # List of fields
    # self.dataset.docs_cls()._fields

    def iter(self) -> Iterator[ir.Document]:
        """Returns an iterator over adhoc documents"""
        for doc in self.dataset.docs_iter():
            yield self.converter(doc)

    @property
    def documentcount(self):
        return self.dataset.docs_count()

    @cached_property
    def converter(self):
        return Documents.CONVERTERS.get(
            self.dataset.docs_cls(), lambda doc: IRDSDocumentWrapper(doc)
        )

    @cached_property
    def store(self):
        return self.dataset.docs_store()

    def docid_internal2external(self, ix: int):
        return self.dataset.docs_iter()[ix].doc_id

    def document_ext(self, docid: str) -> Document:
        return self.converter(self.store.get(docid))

    def document_int(self, ix):
        return self.converter(self.dataset.docs_iter()[ix])

    @cached_property
    def document_cls(self):
        return self.converter.target_cls

    @cached_property
    def converter(self):
        converter = Documents.CONVERTERS[self.dataset.docs_cls()]
        converter.check(self.dataset.docs_cls())
        return converter


@attrs.define()
class IRDSQueryWrapper(ir.Topic):
    query: Any


class Topics(ir.TopicsStore, IRDSId):
    CONVERTERS = {
        GenericQuery: tuple_constructor(GenericTopic, "query_id", "text"),
        _irds.beir.BeirCovidQuery: tuple_constructor(
            formats.TrecTopic, "query_id", "text", "query", "narrative"
        ),
    }

    def iter(self) -> Iterator[ir.Topic]:
        """Returns an iterator over topics"""
        for query in self.dataset.queries_iter():
            yield self.converter(query)

    def count(self):
        return self.dataset.queries_count()

    def topic_int(self, internal_topic_id: int) -> Topic:
        """Returns a document given its internal ID"""
        return self.topics_list[internal_topic_id]

    def topic_ext(self, external_topic_id: int) -> Topic:
        """Returns a document given its external ID"""
        return self.topics_map[external_topic_id]

    @cached_property
    def topics_map(self):
        return self.topics[0]

    @cached_property
    def topics_list(self):
        return self.topics[1]

    @cached_property
    def topics(self):
        topic_map = {}
        topic_list = []
        for query in self.iter():
            topic_list.append(query)
            topic_map[query.get_id()] = query

        return topic_map, topic_list

    @cached_property
    def topic_cls(self):
        return self.converter.target_cls

    @cached_property
    def converter(self):
        converter = Topics.CONVERTERS[self.dataset.queries_cls()]
        converter.check(self.dataset.queries_cls())
        return converter


class Adhoc(ir.Adhoc, IRDSId):
    pass


class AdhocRun(ir.AdhocRun, IRDSId):
    pass


class TrainingTriplets(ir.TrainingTriplets, IRDSId):
    """Training triplets from IR Dataset"""

    CONVERTERS = {
        GenericDocPair: lambda qid, doc1_id, doc2_id: (
            IDTopic(qid),
            IDDocument(doc1_id),
            IDDocument(doc2_id),
        )
    }

    @cached_property
    def topic_cls(self):
        return IDTopic

    @cached_property
    def document_cls(self):
        return IDDocument

    @cached_property
    def converter(self):
        return TrainingTriplets.CONVERTERS[self.dataset.docpairs_cls()]

    def iter(self) -> Iterator[Tuple[ir.Topic, ir.Document, ir.Document]]:
        ds = self.dataset

        logging.info("Starting to generate triplets")
        yield from (self.converter(*entry) for entry in ds.docpairs_iter())
        logging.info("Ending triplet generation")

    def count(self):
        """Returns the length or None"""
        return self.dataset.docpairs_count()



from typing import ClassVar
from attrs import define
from .base import IDHolder, Document, GenericTopic


@define
class CordDocument(IDHolder, Document):
    text: str
    title: str
    url: str
    pubmed_id: str

    has_text: ClassVar[bool] = True

    def get_text(self):
        return f"{self.title} {self.text}"


@define
class TrecTopic(GenericTopic):
    text: str
    query: str
    narrative: str

    def get_text(self):
        return f"{self.query}"



# Copyright (c) 2016 Salle, Alexandre <atsalle@inf.ufrgs.br>
# Author: Salle, Alexandre <atsalle@inf.ufrgs.br>
# Modified: Benjamin Piwowarski <b@piwowarski.fr>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import tempfile
from pathlib import Path
import atexit
import numpy
import logging
from typing import TextIO, Optional, List

# Use 1GB
MEMORY = 1024**3


def shuffle_and_close(random, buf, f):
    random.shuffle(buf)
    f.writelines(buf)
    f.close()


def add_temporary_file(tmp_path: Path, files: List, rm_files: List[str]):
    tmp_file = tempfile.NamedTemporaryFile(mode="w", delete=False, dir=tmp_path)
    logging.info("Adding temporary file %s", tmp_file.name)
    files.append(tmp_file)
    rm_files.append(tmp_file.name)


def shuffle(
    input,
    output: TextIO,
    *,
    memory=MEMORY,
    random=None,
    tmp_path: Optional[Path] = None
):
    """Shuffle using temporary file"""
    if random is None:
        random = numpy.random.RandomState()
    files = []

    # --- Files to remove
    rm_files = []

    def remove_files():
        logging.info("Removing files")
        for path in rm_files:
            if os.path.exists(path):
                os.unlink(path)

    # In case of crash, remove the files
    atexit.register(remove_files)

    # Let's go
    add_temporary_file(tmp_path, files, rm_files)
    total_bytes = 0
    total_lines = 0
    buf = []
    bytes_used = 0

    # Create the temporary files
    for line in input:
        bytes_used += len(line)
        total_bytes += len(line)
        total_lines += 1
        buf.append(line)
        if bytes_used >= memory:
            shuffle_and_close(random, buf, files[-1])
            add_temporary_file(tmp_path, files, rm_files)

            buf = []
            bytes_used = 0

    if buf:
        shuffle_and_close(random, buf, files[-1])

    buf = []
    avg_bytes_per_line = total_bytes / float(total_lines)
    logging.info("Output from %d temporary files", len(files))

    files = [open(f.name) for f in files]
    while files:
        lines_per_file = int((memory / avg_bytes_per_line) / len(files))
        for f in files:
            lines = f.readlines(lines_per_file)
            buf += lines
            if not lines:
                files.remove(f)
                os.unlink(f.name)
                rm_files.remove(f.name)

        random.shuffle(buf)
        output.writelines(buf)
        buf = []

    remove_files()
    atexit.unregister(remove_files)



"""MS MARCO (Microsoft Machine Reading Comprehension) is a large scale dataset focused on machine reading comprehension, question answering, and passage ranking. A variant of this task will be the part of TREC and AFIRM 2019. For Updates about TREC 2019 please follow This Repository Passage Reranking task Task Given a query q and a the 1000 most relevant passages P = p1, p2, p3,... p1000, as retrieved by BM25 a succeful system is expected to rerank the most relevant passage as high as possible. For this task not all 1000 relevant items have a human labeled relevant passage. Evaluation will be done using MRR.

  **Publication**:
  Tri Nguyen, Mir Rosenberg, Xia Song, Jianfeng Gao, Saurabh Tiwary, RanganMajumder, and Li Deng. 2016.
  MS MARCO: A Human Generated MAchineReading COmprehension Dataset. In CoCo@NIPS.


  See [https://github.com/microsoft/MSMARCO-Passage-Ranking](https://github.com/microsoft/MSMARCO-Passage-Ranking) for more details
"""

from datamaestro.annotations.agreement import useragreement
from datamaestro.data import Folder
from datamaestro.download.single import filedownloader
from datamaestro.download import reference
from datamaestro.definitions import datatasks, datatags, dataset
from datamaestro.download.archive import tardownloader
from datamaestro_text.data.ir import RerankAdhoc, Adhoc
from datamaestro_text.data.ir.csv import (
    Topics,
    AdhocRunWithText,
    Documents,
    TrainingTriplets,
    TrainingTripletsID,
)
from datamaestro_text.data.ir.trec import TrecAdhocAssessments
from datamaestro.utils import HashCheck
from hashlib import md5


# User agreement
lua = useragreement(
    """Will begin downloading MS-MARCO dataset.
Please confirm you agree to the authors' data usage stipulations found at
http://www.msmarco.org/dataset.aspx""",
    id="net.windows.msmarco",
)

# --- Document collection

# TODO: Not ideal since it would be better to have small versions right away
# instead of downloading again the MS Marco Collection
@lua
@tardownloader(
    "data",
    url="https://msmarco.blob.core.windows.net/msmarcoranking/collectionandqueries.tar.gz",
    checker=HashCheck("31644046b18952c1386cd4564ba2ae69", md5),
)
@dataset(Folder, url="https://github.com/microsoft/MSMARCO-Passage-Ranking")
def collection_etc(data):
    """Documents and some more files"""
    return {"path": data}


@lua
@reference("data", collection_etc)
@dataset(Documents, size="2.9GB")
def collection(data):
    """MS-Marco documents

    This file contains each passage in the larger MSMARCO dataset.

    Format is TSV (PID \t Passage)"""
    return {"path": data.path / "collection.tsv"}


# --- Train


@lua
@tardownloader(
    "run",
    url="https://msmarco.blob.core.windows.net/msmarcoranking/top1000.train.tar.gz",
    checker=HashCheck("d99fdbd5b2ea84af8aa23194a3263052", md5),
)
@dataset(AdhocRunWithText, size="2.5GB")
def train_run(run):
    """

    TSV format: qid, pid, query, passage
    """
    return {"path": run / "top1000.train.tsv"}


@lua
@tardownloader(
    "queries",
    url="https://msmarco.blob.core.windows.net/msmarcoranking/queries.tar.gz",
    files=["queries.train.tsv"],
    checker=HashCheck("c177b2795d5f2dcc524cf00fcd973be1", md5),
)
@dataset(Topics)
def train_queries(queries):
    return {"path": queries / "queries.train.tsv"}


@lua
@filedownloader(
    "qrels.tsv",
    url="https://msmarco.blob.core.windows.net/msmarcoranking/qrels.train.tsv",
    checker=HashCheck("733fb9fe12d93e497f7289409316eccf", md5),
)
@dataset(TrecAdhocAssessments, size="10.1MB")
def train_qrels(qrels):
    return {"path": qrels}


@lua
@reference("collection", collection)
@reference("topics", train_queries)
@reference("qrels", train_qrels)
@datatasks("information retrieval", "passage retrieval")
@dataset(Adhoc, url="https://github.com/microsoft/MSMARCO-Passage-Ranking")
def train(topics, qrels, collection):
    """MS-Marco train dataset"""
    return {
        "documents": collection,
        "topics": topics,
        "assessments": qrels,
    }


@lua
@reference("train", train)
@reference("run", train_run)
@datatasks("information retrieval", "passage retrieval")
@dataset(RerankAdhoc, url="https://github.com/microsoft/MSMARCO-Passage-Ranking")
def train_withrun(train, run):
    """MSMarco train dataset, including the top-1000 to documents to re-rank"""
    return {**train.__arguments__(), "run": run}


# Training triplets


@filedownloader(
    "triples.tsv",
    size=1_841_693_309,
    url="https://msmarco.blob.core.windows.net/msmarcoranking/qidpidtriples.train.full.2.tsv.gz",
    checker=HashCheck("4e58f45f82f3fe99e3239ecffd8ed371", md5),
)
@reference("collection", collection)
@reference("topics", train_queries)
@dataset(
    TrainingTripletsID,
    url="https://github.com/microsoft/MSMARCO-Passage-Ranking",
    size="5.7GB",
)
def train_idtriples(topics, collection, triples):
    """Full training triples (query, positive passage, negative passage) with IDs"""
    return {"path": triples, "topics": topics, "documents": collection}


@filedownloader(
    "triples.tsv",
    size=7_930_881_353,
    url="https://msmarco.blob.core.windows.net/msmarcoranking/triples.train.small.tar.gz",
    checker=HashCheck("c13bf99ff23ca691105ad12eab837f84", md5),
)
@dataset(
    TrainingTriplets,
    url="https://github.com/microsoft/MSMARCO-Passage-Ranking",
    size="27.1GB",
)
def train_texttriples_small(triples):
    """Small training triples (query, positive passage, negative passage) with text"""
    return {"path": triples}


@filedownloader(
    "triples.tsv",
    size=77_877_731_328,
    url="https://msmarco.blob.core.windows.net/msmarcoranking/triples.train.full.tar.gz",
    checker=HashCheck("8d509d484ea1971e792b812ae4800c6f", md5),
)
@dataset(
    TrainingTriplets,
    url="https://github.com/microsoft/MSMARCO-Passage-Ranking",
    size="272.2GB",
)
def train_texttriples_full(triples):
    """Full training triples (query, positive passage, negative passage) with text"""
    return {"path": triples}


# ---
# --- Development set
# ---


@lua
@tardownloader(
    "queries",
    url="https://msmarco.blob.core.windows.net/msmarcoranking/queries.tar.gz",
    files=["queries.dev.tsv"],
    checker=HashCheck("c177b2795d5f2dcc524cf00fcd973be1", md5),
)
@dataset(Topics)
def dev_queries(queries):
    return {"path": queries / "queries.dev.tsv"}


@lua
@tardownloader(
    "run",
    url="https://msmarco.blob.core.windows.net/msmarcoranking/top1000.dev.tar.gz",
    checker=HashCheck("8c140662bdf123a98fbfe3bb174c5831", md5),
)
@dataset(AdhocRunWithText)
def dev_run(run):
    return {"path": run / "top1000.eval.tsv"}


@lua
@filedownloader(
    "qrels.tsv",
    url="https://msmarco.blob.core.windows.net/msmarcoranking/qrels.dev.tsv",
    checker=HashCheck("9157ccaeaa8227f91722ba5770787b16", md5),
)
@dataset(TrecAdhocAssessments)
def dev_qrels(qrels):
    return {"path": qrels}


@lua
@reference("collection", collection)
@reference("topics", dev_queries)
@reference("qrels", dev_qrels)
@datatasks("information retrieval", "passage retrieval")
@dataset(Adhoc, url="https://github.com/microsoft/MSMARCO-Passage-Ranking")
def dev(topics, qrels, collection):
    """MS-Marco dev dataset"""
    return {
        "documents": collection,
        "topics": topics,
        "assessments": qrels,
    }


@lua
@reference("dev", dev)
@reference("run", dev_run)
@datatasks("information retrieval", "passage retrieval")
@dataset(RerankAdhoc, url="https://github.com/microsoft/MSMARCO-Passage-Ranking")
def dev_withrun(dev, run):
    """MSMarco dev dataset, including the top-1000 to documents to re-rank"""
    return {**dev.__arguments__(), "run": run}


@lua
@tardownloader(
    "run",
    url="https://msmarco.blob.core.windows.net/msmarcoranking/top1000.eval.tar.gz",
    checker=HashCheck("73778cd99f6e0632d12d0b5731b20a02", md5),
)
@dataset(AdhocRunWithText)
def eval_withrun(run):
    return {"path": run / "top1000.eval.tsv"}


# ---
# --- Relevant Passages
# --- https://github.com/microsoft/MSMARCO-Passage-Ranking#relevant-passages
# ---


@reference("data", collection_etc)
@dataset(Topics, url="https://github.com/microsoft/MSMARCO-Passage-Ranking")
def dev_small_queries(data):
    return {"path": data.path / "queries.dev.small.tsv"}


@reference("data", collection_etc)
@dataset(
    TrecAdhocAssessments, url="https://github.com/microsoft/MSMARCO-Passage-Ranking"
)
def dev_small_qrels(data):
    return {"path": data.path / "qrels.dev.small.tsv"}


@reference("topics", dev_small_queries)
@reference("qrels", dev_small_qrels)
@reference("collection", collection)
@dataset(Adhoc, url="https://github.com/microsoft/MSMARCO-Passage-Ranking")
def dev_small(collection, topics, qrels):
    return {"documents": collection, "topics": topics, "assessments": qrels}


@reference("data", collection_etc)
@dataset(Topics, url="https://github.com/microsoft/MSMARCO-Passage-Ranking")
def eval_queries_small(data):
    return {"path": data.path / "queries.eval.small.tsv"}


# ---
# --- TREC 2019
# ---


@lua
@filedownloader(
    "queries.tsv",
    url="https://msmarco.blob.core.windows.net/msmarcoranking/msmarco-test2019-queries.tsv.gz",
    checker=HashCheck("756e60d714cee28d3b552289d6272f1d", md5),
)
@dataset(Topics)
def trec2019_test_queries(queries):
    return {"path": queries}


@lua
@filedownloader(
    "run.tsv",
    url="https://msmarco.blob.core.windows.net/msmarcoranking/msmarco-passagetest2019-top1000.tsv.gz",
    checker=HashCheck("ec9e012746aa9763c7ff10b3336a3ce1", md5),
)
@dataset(AdhocRunWithText)
def trec2019_test_run(run):
    return {"path": run / "top1000.eval.tsv"}


@lua
@filedownloader(
    "qrels.tsv",
    url="https://trec.nist.gov/data/deep/2019qrels-pass.txt",
    checker=HashCheck("2f4be390198da108f6845c822e5ada14", md5),
)
@dataset(TrecAdhocAssessments)
def trec2019_test_qrels(qrels):
    return {"path": qrels}


@lua
@reference("collection", collection)
@reference("topics", trec2019_test_queries)
@reference("qrels", trec2019_test_qrels)
@datatasks("information retrieval", "passage retrieval")
@dataset(Adhoc, url="https://microsoft.github.io/msmarco/TREC-Deep-Learning-2019.html")
def trec2019_test(topics, qrels, collection):
    "TREC Deep Learning (2019)"
    return {
        "documents": collection,
        "topics": topics,
        "assessments": qrels,
    }


@lua
@reference("trec2019", trec2019_test)
@reference("run", trec2019_test_run)
@datatasks("information retrieval", "passage retrieval")
@dataset(
    RerankAdhoc, url="https://microsoft.github.io/msmarco/TREC-Deep-Learning-2019.html"
)
def trec2019_test_withrun(trec2019, run):
    """TREC Deep Learning (2019), including the top-1000 to documents to re-rank"""
    return {**trec2019.__arguments__(), "run": run}


# ---
# --- TREC 2020
# ---


@lua
@filedownloader(
    "queries.tsv",
    url="https://msmarco.blob.core.windows.net/msmarcoranking/msmarco-test2020-queries.tsv.gz",
    checker=HashCheck("00a406fb0d14ed3752d70d1e4eb98600", md5),
)
@dataset(Topics, size="12K")
def trec2020_test_queries(queries):
    """TREC Deep Learning 2019 (topics)

    Topics of the TREC 2019 MS-Marco Deep Learning track"""
    return {"path": queries}


@lua
@datatasks("information retrieval", "passage retrieval")
@datatags("reranking")
@filedownloader(
    "run.tsv",
    url="https://msmarco.blob.core.windows.net/msmarcoranking/msmarco-passagetest2020-top1000.tsv.gz",
    checker=HashCheck("aa6fbc51d66bd1dc745964c0e140a727", md5),
)
@dataset(
    AdhocRunWithText,
    url="https://microsoft.github.io/msmarco/TREC-Deep-Learning-2020.html",
)
def trec2020_test_run(run):
    """TREC Deep Learning (2020)

    Set of query/passages for the passage re-ranking task re-rank (TREC 2020)"""
    return {"path": run / "top1000.eval.tsv"}



"""

TIPSTER is sometimes also called the Text Research Collection Volume or TREC.

The TIPSTER project was sponsored by the Software and Intelligent Systems Technology
Office of the Advanced Research Projects Agency (ARPA/SISTO) in an effort to significantly
advance the state of the art in effective document detection (information retrieval) and
data extraction from large, real-world data collections.

The detection data is comprised of a test collection built at NIST for the TIPSTER project
and the related TREC project. The TREC project has many other participating information
retrieval research groups, working on the same task as the TIPSTER groups, but meeting
once a year in a workshop to compare results (similar to MUC). The test collection consists
of three CD-ROMs of SGML encoded documents distributed by LDC plus queries and answers
(relevant documents) distributed by NIST.

See also https://trec.nist.gov/data/docs_eng.html and https://trec.nist.gov/data/intro_eng.html
"""

from datamaestro.data import Base
from datamaestro_text.data.ir.trec import TipsterCollection
from datamaestro.download.links import linkfolder
from datamaestro.definitions import (
    dataset,
    DatafolderPath,
)

# Store meta-information
TIPSTER = dataset(TipsterCollection, url="https://catalog.ldc.upenn.edu/LDC93T3A")


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk2/AP")])
@TIPSTER
def ap88(documents):
    """Associated Press document collection (1988)"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk1/AP")])
@TIPSTER
def ap89(documents):
    """Associated Press document collection (1989)"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk3/AP")])
@TIPSTER
def ap90(documents):
    """Associated Press document collection (1990)"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk1/DOE")])
@TIPSTER
def doe1(documents):
    """Department of Energy documents"""
    return {"path": documents}


# --- Wall Street Journal (1987-92)


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk1/WSJ/1987")])
@TIPSTER
def wsj87(documents):
    """Wall Street Journal (1987)"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk1/WSJ/1988")])
@TIPSTER
def wsj88(documents):
    """Wall Street Journal (1988)"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk1/WSJ/1989")])
@TIPSTER
def wsj89(documents):
    """Wall Street Journal (1989)"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk2/WSJ/1990")])
@TIPSTER
def wsj90(documents):
    """Wall Street Journal (1990)"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk2/WSJ/1991")])
@TIPSTER
def wsj91(documents):
    """Wall Street Journal (1991)"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk2/WSJ/1992")])
@TIPSTER
def wsj92(documents):
    """Wall Street Journal (1992)"""
    return {"path": documents}


# --- Federal Register (1988-89)


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk2/FR")])
@TIPSTER
def fr88(documents):
    """Federal Register (1988)"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk1/FR")])
@TIPSTER
def fr89(documents):
    """Federal Register (1989)"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk4/FR94")])
@TIPSTER
def fr94(documents):
    """Federal Register (1994)"""
    return {"path": documents}


# # ZIFF (1988-92)


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk1/ZIFF")])
@TIPSTER
def ziff1(documents):
    """Information from the Computer Select disks (1989-90)"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk2/ZIFF")])
@TIPSTER
def ziff2(documents):
    """Information from the Computer Select disks (1989-90)"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk3/ZIFF")])
@TIPSTER
def ziff3(documents):
    """Information from the Computer Select disks (1990-91)"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk3/SJM")])
@TIPSTER
def sjm1(documents):
    """San Jose Mercury News (1991)"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk4/CR")])
@TIPSTER
def cr1(documents):
    """TODO"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk4/FT")])
@TIPSTER
def ft1(documents):
    """Financial Times"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk5/FBIS")])
@TIPSTER
def fbis1(documents):
    """Foreign Broadcast Information Service (1996)"""
    return {"path": documents}


@linkfolder("documents", [DatafolderPath("gov.nist.trec.tipster", "Disk5/LATIMES")])
@TIPSTER
def la8990(documents):
    """Los Angeles Times (1989-90)"""
    return {"path": documents}



import logging
import gzip
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Type
from experimaestro import Config, Task, Param, Annotated, pathgenerator, Option, tqdm
import numpy as np
import datamaestro_text.data.ir as ir
from datamaestro_text.utils.shuffle import shuffle


def getpathname(context, config):
    name = "triplets.lst"
    if config.compressed:
        name = "triplets.lst.gz"

    return context.currentpath() / name


class StoreTrainingTripletTopicAdapter(ir.TrainingTriplets):
    """Retrieve an adhoc topic from a topic store and returns triplets with those"""

    id: Param[str] = ""

    store: Param[ir.TopicsStore]
    """The topic store to use"""

    data: Param[ir.TrainingTriplets]
    """Input data"""

    def __validate__(self):
        assert self.data.topic_cls.has_id, "Topics have no ID"

    def iter(self):
        for topic, doc1, doc2 in self.data.iter():
            yield self.store.topic_ext(topic.get_id()), doc1, doc2

    def count(self):
        return self.data.count()

    @property
    def topic_cls(self) -> Type[ir.Topic]:
        """The class for topics"""
        return self.store.topic_cls

    @property
    def document_cls(self) -> Type[ir.Document]:
        """The class for documents"""
        return self.data.document_cls


class StoreTrainingTripletDocumentAdapter(ir.TrainingTriplets):
    """Retrieve an adhoc topic from a topic store and returns triplets with those"""

    id: Param[str] = ""

    store: Param[ir.DocumentStore]
    """The topic store to use"""

    data: Param[ir.TrainingTriplets]
    """Input data"""

    def __validate__(self):
        assert self.data.document_cls.has_id, "Documents have no ID"

    def iter(self):
        for topic, doc1, doc2 in self.data.iter():
            yield topic, self.store.document_ext(
                doc1.get_id()
            ), self.store.document_ext(doc2.get_id())

    def count(self):
        return self.data.count()

    @property
    def topic_cls(self) -> Type[ir.Topic]:
        """The class for topics"""
        return self.store.topic_cls

    @property
    def document_cls(self) -> Type[ir.Document]:
        """The class for documents"""
        return self.data.document_cls


class ShuffledTrainingTripletsLines(Task):
    """Shuffle a set of training triplets"""

    data: Param[ir.TrainingTriplets]
    """Input data"""

    path: Annotated[Path, pathgenerator(getpathname)]
    """Output path"""

    doc_ids: Param[bool]
    """Whether to use document ids"""

    topic_ids: Param[bool]
    """True if we have query IDs"""

    seed: Param[int]
    """The random seed"""

    compressed: Option[bool] = True
    """Compress the output"""

    sample_rate: Param[float] = 1.0
    """Sampling rate - set to 1 to keep all the samples"""

    sample_max: Param[int] = 0
    """Maximum number of samples"""

    tmp_path: Annotated[Path, pathgenerator("tmp")]
    """Path where temporary files will be stored"""

    def __validate__(self):
        if self.topic_ids:
            assert self.data.topic_cls.has_id, "No topic ID in the source data"
        else:
            assert self.data.topic_cls.has_text, "No topic text in the source data"

        if self.doc_ids:
            assert self.data.document_cls.has_id, "No doc ID in the source data"
        else:
            assert self.data.document_cls.has_text, "No doc text in the source data"

    def task_outputs(self, dep):
        return dep(
            ir.TrainingTripletsLines(
                id="",
                path=self.path,
                topic_ids=self.topic_ids,
                doc_ids=self.doc_ids,
                sep="\t",
            )
        )

    def execute(self):
        # --- Shuffle using the shuf command with a seed

        random = np.random.RandomState(self.seed)

        if self.topic_ids:

            def get_query(query):
                return query.get_id()

        else:

            def get_query(query):
                return query.get_text()

        if self.doc_ids:

            def get_doc(doc):
                return doc.get_id()

        else:

            def get_doc(doc):
                return doc.get_text()

        def triplegenerator():
            logging.info("Starting to output triples")
            count = 0

            total = self.data.count()
            if self.sample_max > 0:
                total = min(total, self.sample_max)

            pbar = tqdm(total=total)
            for query, doca, docb in self.data.iter():
                # Discard sample
                if self.sample_rate < 1:
                    if random.uniform() > self.sample_rate:
                        continue

                pbar.update(1)
                count += 1
                yield f"{get_query(query)}\t{get_doc(doca)}\t{get_doc(docb)}\n"

                if self.sample_max > 0 and count >= self.sample_max:
                    break

            logging.info("Triples output ended (%d triples)", count)

        logging.info("Creating generator")

        # Output can be a stream or nothing
        if self.compressed:
            output = gzip.open(self.path, "wt")
        else:
            output = self.path.open("wt")

        with output:
            self.tmp_path.mkdir(exist_ok=True)
            shuffle(triplegenerator(), output, random=random, tmp_path=self.tmp_path)


class TopicWrapper(Config, ABC):
    """Modify topics on the fly using a topic wrapper"""

    @abstractmethod
    def __call__(topic: ir.Topic) -> ir.Topic:
        """Transforms a topic into another topic"""
        ...



# See documentation on https://datamaestro.readthedocs.io

from datamaestro.definitions import datatasks, datatags, dataset
from datamaestro.download.archive import zipdownloader
from datamaestro.utils import HashCheck

from datamaestro.data.ml import Supervised
from datamaestro_text.data.conversation.canard import CanardDataset


@datatags("conversation", "context", "query")
@datatasks("query rewriting")
@zipdownloader(
    "archive",
    "https://obj.umiacs.umd.edu/elgohary/CANARD_Release.zip",
    subpath="CANARD_Release",
    checker=HashCheck("c9bba7c6bb898f669383415b54fd6ffd"),
)
@dataset(Supervised, url="https://sites.google.com/view/qanta/projects/canard", id="")
def main(archive):
    """Question-in-context rewriting

    CANARD is a dataset for question-in-context rewriting that consists of
    questions each given in a dialog context together with a context-independent
    rewriting of the question. The context of each question is the dialog
    utterances that precede the question. CANARD can be used to evaluate
    question rewriting models that handle important linguistic phenomena such as
    co-reference and ellipsis resolution.

    Each dataset is an instance of :class:`datamaestro_text.data.conversation.CanardDataset`
    """
    return {
        "train": CanardDataset(path=archive / "train.json"),
        "validation": CanardDataset(path=archive / "dev.json"),
        "test": CanardDataset(path=archive / "test.json"),
    }



# IR-datasets data types

# https://ir-datasets.com/


import datamaestro


class Repository(datamaestro.Repository):
    NAMESPACE = "irds"
    DESCRIPTION = (
        "Information Retrieval datasets repository from https://ir-datasets.com/"
    )

    def __init__(self, context):
        super().__init__(context)
        self._modules = None
        self._datasets = None

    def _check(self):
        from .datasets import build

        if self._modules is None:
            self._modules, self._datasets = build(self)

    def modules(self):
        self._check()
        yield from self._modules

    def search(self, name: str):
        self._check()
        return self._datasets.get(name, None)



from ir_datasets.formats import GenericQuery, GenericDoc


def get_text(object) -> str:
    """Returns the text of an object (e.g. query, document)

    :param query: any object
    """
    if isinstance(object, GenericQuery, GenericDoc):
        return object.text

    raise NotImplementedError(
        f"No pre-defined method to retrieve the text of an object of type {type(object)}"
    )


def get_qid_text(object) -> str:
    """Returns the ID and the text of an object (e.g. query, document)

    :param query: any object
    """
    if isinstance(object, GenericDoc):
        return object.doc_id, object.text

    if isinstance(object, GenericQuery):
        return object.query_id, object.text

    raise NotImplementedError(
        f"No pre-defined method to retrieve the text of an object of type {type(object)}"
    )



# See documentation on https://datamaestro.readthedocs.io

from datamaestro.definitions import datatasks, datatags, dataset
from datamaestro.download.single import filedownloader
from datamaestro.utils import HashCheck


from datamaestro_text.data.conversation.orconvqa import OrConvQADataset
from datamaestro.data.ml import Supervised


@datatags("conversation", "context", "query")
@datatasks("query rewriting")
@filedownloader(
    "train.jsonl",
    "https://ciir.cs.umass.edu/downloads/ORConvQA/preprocessed/train.txt",
    checker=HashCheck("7513a9ef12d8b7a4471166dc4fef77b7"),
)
@filedownloader(
    "dev.jsonl",
    "https://ciir.cs.umass.edu/downloads/ORConvQA/preprocessed/dev.txt",
    checker=HashCheck("7765658995cc9ffd5eb39a400d814b20"),
)
@filedownloader(
    "test.jsonl",
    "https://ciir.cs.umass.edu/downloads/ORConvQA/preprocessed/test.txt",
    checker=HashCheck("0cf3a755f06297b9c02e7db45f8dc8be"),
)
@dataset(
    Supervised,
    url="https://github.com/prdwb/orconvqa-release",
)
def preprocessed(train, dev, test):
    """Question-in-context rewriting

    CANARD is a dataset for question-in-context rewriting that consists of
    questions each given in a dialog context together with a context-independent
    rewriting of the question. The context of each question is the dialog
    utterances that precede the question. CANARD can be used to evaluate
    question rewriting models that handle important linguistic phenomena such as
    co-reference and ellipsis resolution.

    Each dataset is an instance of :class:`datamaestro_text.data.conversation.OrConvQADataset`
    """
    return {
        "train": OrConvQADataset(path=train),
        "validation": OrConvQADataset(path=dev),
        "test": OrConvQADataset(path=test),
    }



from typing import Iterator, List, Optional
from attr import define
import json
from datamaestro.data import File

from .base import (
    AnswerEntry,
    Conversation,
    DecontextualizedEntry,
    RetrievedEntry,
    Conversation,
)
from . import ConversationDataset


@define(kw_only=True)
class OrConvQADatasetAnswer:
    text: staticmethod
    answer_start: 0
    bid: Optional[int] = None


@define(kw_only=True)
class OrConvQADatasetHistoryEntry:
    question: str
    """The question"""

    answer: OrConvQADatasetAnswer


@define(kw_only=True)
class OrConvQADatasetEntry:
    """A query with past history"""

    query_id: int
    """Question number"""

    query: str
    """The last issued query"""

    rewrite: str
    """Manually rewritten query"""

    history: List[OrConvQADatasetHistoryEntry]
    """The list of queries asked by the user"""

    answer: OrConvQADatasetAnswer

    evidences: List[str]
    """Evidence sources for the conversation"""

    retrieval_labels: List[int]
    """Relevance status for evidences"""


@define(kw_only=True, slots=False)
class OrConvQAEntry(DecontextualizedEntry, RetrievedEntry, AnswerEntry):
    """Entry"""


class OrConvQADataset(ConversationDataset[OrConvQAEntry], File):
    def iter(self) -> Iterator[OrConvQADatasetEntry]:
        """Iterates over re-written query with their context"""
        with self.path.open("rt") as fp:
            for line in fp.readlines():
                entry = json.loads(line)
                yield OrConvQADatasetEntry(
                    query_id=entry["qid"],
                    query=entry["question"],
                    evidences=entry["evidences"],
                    retrieval_labels=entry["retrieval_labels"],
                    rewrite=entry["rewrite"],
                    answer=OrConvQADatasetAnswer(**entry["answer"]),
                    history=[
                        OrConvQADatasetHistoryEntry(
                            question=history_entry["question"],
                            answer=OrConvQADatasetAnswer(**history_entry["answer"]),
                        )
                        for history_entry in entry["history"]
                    ],
                )

    def iter_conversations(self) -> Iterator[Conversation[OrConvQAEntry]]:
        current: Optional[Conversation] = None

        for entry in self.iter():
            # Check if current conversation
            cid, query_no = entry.query_id.rsplit("#", 1)
            if current is None or cid != current.id:
                if current is not None:
                    yield current
                current = Conversation(id=cid, history=[])

            # Add to current
            current.history.append(
                OrConvQAEntry(
                    query=entry.query,
                    decontextualized_query=entry.rewrite,
                    answer=entry.answer.text,
                    documents=entry.evidences,
                    document_relevances=entry.retrieval_labels,
                )
            )

        yield current



import tempfile
import os
import random
import logging
from pathlib import Path
from threading import Thread


class RandomStream(Thread):
    """Generate a FIFO file with random bytes"""

    def __init__(self, seed: int):
        super().__init__()
        tmpdir = tempfile.mkdtemp()
        self.filepath = Path(tmpdir) / "random"
        os.mkfifo(self.filepath)
        self.random = random.Random(seed)
        self.thread = None

    def run(self):
        logging.info("Starting to write random bytes %s", self.filepath)
        try:
            with self.filepath.open("wb") as out:
                while True:
                    data = bytearray(random.getrandbits(8) for _ in range(256))
                    out.write(data)
        except BrokenPipeError:
            pass

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.join()
        self.filepath.unlink()
        self.filepath.parent.rmdir()



from pathlib import Path
from typing import Iterator, Tuple

from experimaestro import Param, Option, Constant, Meta
from datamaestro.definitions import argument
import datamaestro_text.data.ir as ir
from datamaestro_text.data.ir.base import GenericTopic
from datamaestro_text.interfaces.plaintext import read_tsv


@argument("path", type=Path)
@argument("separator", type=str, default="\t", ignored=True)
class AdhocRunWithText(ir.AdhocRun):
    "(qid, doc.id, query, passage)"
    pass


@argument("path", type=Path)
@argument("separator", type=str, default="\t", ignored=True)
class Topics(ir.Topics):
    "Pairs of query id - query using a separator"

    def iter(self):
        return (GenericTopic(qid, title) for qid, title in read_tsv(self.path))


class Documents(ir.Documents):
    "One line per document, format pid<SEP>text"

    path: Param[Path]
    separator: Meta[str] = "\t"

    pass


class TrainingTripletsID(ir.TrainingTripletsLines):
    """Training triplets (query/document IDs only)

    Attributes:
        separator: Field separator
        documents: The documents
        topics: The topics
        ids: Whether documents are IDs or full text
    """

    separator: Option[str] = "\t"
    documents: Param[ir.Documents]
    topics: Param[ir.Topics]
    ids: Constant[bool] = True

    def iter(self) -> Iterator[Tuple[str, str, str]]:
        queries = {}
        for query in self.topics.iter():
            queries[query.get_id()] = query.get_text()

        for qid, pos, neg in read_tsv(self.path):
            yield queries[qid], pos, neg


class TrainingTriplets(ir.TrainingTriplets):
    "Training triplets (full text)"
    path: Param[Path]
    separator: Meta[str] = "\t"
    ids: Constant[bool] = True

    def iter(self) -> Iterator[Tuple[str, str, str]]:
        yield from read_tsv(self.path)



from pathlib import Path
from datamaestro.data import Base, Folder, File, argument
from datamaestro.data.ml import Supervised


@argument("train", type=Base)
@argument("test", type=Base, required=False)
@argument("validation", type=Base, required=False)
class TrainingText(Supervised):
    """ "A dataset used for training with a train and a test"""

    pass


class TextFolder(Folder):
    "A folder composed of texts"
    pass


class TextFile(File):
    """A file composed of texts"""

    pass



from datamaestro.data.csv import Generic
from datamaestro.definitions import argument, datatasks, datatags, dataset
from datamaestro.download.archive import zipdownloader
from datamaestro.data.ml import Supervised
from datamaestro.utils import HashCheck

# name: Sentiment140
# web: http://help.sentiment140.com/for-students/

# description: |


@zipdownloader(
    "dir",
    "http://cs.stanford.edu/people/alecmgo/trainingandtestdata.zip",
    checker=HashCheck("1647eb110dd2492512e27b9a70d5d1bc"),
)
@datatasks("sentiment analysis")
@datatags("english", "sentiment", "text")
@dataset(Supervised, url="http://help.sentiment140.com/for-students/", size="228M")
def english(dir):
    """Sentiment analysis dataset 140

    The data is a CSV with emoticons removed. Data file format has 6 fields:
    0 - the polarity of the tweet (0 = negative, 2 = neutral, 4 = positive)
    1 - the id of the tweet (2087)
    2 - the date of the tweet (Sat May 16 23:58:44 UTC 2009)
    3 - the query (lyx). If there is no query, then this value is NO_QUERY.
    4 - the user that tweeted (robotickilldozr)
    5 - the text of the tweet (Lyx is cool)

    If you use this data, please cite Sentiment140 as your source.
    """
    return {
        "train": Generic(path=dir / "training.1600000.processed.noemoticon.csv"),
        "test": Generic(path=dir / "testdata.manual.2009.06.14.csv"),
    }



# tasks:
#   - Recommendation
#   - Collaborative Filtering

# download:
#   handler: /archive:Zip
#   url: http://files.grouplens.org/datasets/movielens/ml-20m.zip
#   size: 190M
#   checksum: cd245b17a1ae2cc31bb14903e1204af3
# ...
# ---
# id: tmdb
# description: TMDB (The Movie database) download for MovieLens movies
# download:
#   handler: tmdb:MovieLens


from datamaestro.definitions import dataset
from datamaestro.download.archive import zipdownloader
import datamaestro.data.csv as csv
from datamaestro_text.data.recommendation import Movielens


@zipdownloader(
    "ds", "http://files.grouplens.org/datasets/movielens/ml-latest-small.zip"
)
@dataset(url="https://grouplens.org/datasets/movielens/latest/", timestamp=True)
def small(ds) -> Movielens:
    """MovieLens (small dataset)

    100,000 ratings and 3,600 tag applications applied to 9,000 movies by 600 users (as of 9/2018)
    """
    return {
        "ratings": csv.Generic(path=ds / "ratings.csv", names_row=0),
        "links": csv.Generic(path=ds / "links.csv", names_row=0),
        "movies": csv.Generic(path=ds / "movies.csv", names_row=0),
        "tags": csv.Generic(path=ds / "tags.csv", names_row=0),
    }


@zipdownloader("ds", "http://files.grouplens.org/datasets/movielens/ml-latest.zip")
@dataset(url="https://grouplens.org/datasets/movielens/latest/", timestamp=True)
def full(ds) -> Movielens:
    """MovieLens (full dataset)

    27,000,000 ratings and 1,100,000 tag applications applied to 58,000 movies by 280,000 users (as of 9/2018)
    """
    return {
        "ratings": csv.Generic(path=ds / "ratings.csv", names_row=0),
        "links": csv.Generic(path=ds / "links.csv", names_row=0),
        "movies": csv.Generic(path=ds / "movies.csv", names_row=0),
        "tags": csv.Generic(path=ds / "tags.csv", names_row=0),
    }



from typing import Iterator, List, Optional
from attr import define
import json
from datamaestro.data import File
from .base import Conversation, DecontextualizedEntry
from . import ConversationDataset


@define(kw_only=True)
class CanardConversation:
    """A query with past history"""

    history: List[str]
    """The list of queries asked by the user"""

    query: str
    """The last issued query"""

    rewrite: str
    """Manually rewritten query"""

    dialogue_id: str
    """Conversation identifier"""

    query_no: int
    """Question number"""


@define(kw_only=True, slots=False)
class CanardEntry(DecontextualizedEntry):
    """Entry"""


class CanardDataset(ConversationDataset, File):
    """A dataset in the CANARD json format"""

    def iter(self) -> Iterator[CanardConversation]:
        """Iterates over re-written query with their context"""
        with self.path.open("rt") as fp:
            data = json.load(fp)

        for entry in data:
            yield CanardConversation(
                history=entry["History"],
                query=entry["Question"],
                rewrite=entry["Rewrite"],
                dialogue_id=entry["QuAC_dialog_id"],
                query_no=entry["Question_no"],
            )

    def iter_conversations(self) -> Iterator[Conversation[CanardEntry]]:
        current: Optional[Conversation] = None

        for entry in self.iter():
            # Check if current conversation
            cid = entry.dialogue_id
            if current is None or cid != current.id:
                if current is not None:
                    yield current
                current = Conversation(id=cid, history=[])

            # Add to current
            current.history.append(
                CanardEntry(
                    query=entry.query,
                    decontextualized_query=entry.rewrite,
                )
            )

        yield current



# --- Task and types definitions

import unittest

from datamaestro.test import DatasetTests


class MainTest(DatasetTests, unittest.TestCase):
    pass


if __name__ == "__main__":
    unittest.main()
